from datetime import date, timedelta

from .base import CERT, APIBase
from .exceptions import raises_from_ebs as raises_ethel_exception
from .utils import Template


class RegnumV5(APIBase):
    PAYLOAD_TEMPLATE = Template("payloads/pool.yml")

    def __init__(self, api_host: str) -> None:
        """Subscription registration API

        Access the API for creating subscription pools.

        Args:
            api_host (str): Base API host
        """
        super().__init__(f"https://subscription.{api_host}/svcrest/regnum/v5", cert=CERT)

    @raises_ethel_exception
    def order(
        self,
        username: str,
        sku_id: str,
        quantity: int,
        start_date: date,
        duration: timedelta,
    ) -> dict:
        """Request a subscription pool to be created.

        Args:
            username (str): Account's username.
            sku_id (str): Subscription SKU identifier.
            quantity (int): Pool quantity.
            start_date (date): Date when the pool should become active.
            duration (timedelta): How long the pool should last.

        Returns:
            dict: Information about the placed order.
        """
        payload = self.PAYLOAD_TEMPLATE.render(
            username=username,
            sku_id=sku_id,
            quantity=quantity,
            start_date=start_date,
            duration=duration.days,
        )

        response = self.api.put("/hock/order", json=payload)
        response.raise_for_status()
        return response.json()


class ActivationV2(APIBase):
    def __init__(self, api_host: str) -> None:
        """Subscription activation API

        Access the API for activating subscription pools in a organization.

        Args:
            api_host (str): Base API host
        """
        super().__init__(
            f"https://subscription.{api_host}/svcrest/activation/v2", cert=CERT
        )

    @raises_ethel_exception
    def activate(self, username: str, org_id: int, regnum: int, start_date: date) -> dict:
        """Activate a pool by a registration number.

        Args:
            username (str): Account's username (pool owner).
            org_id (int): Organization ID (pool owner).
            regnum (int): Registration number from the order system.
            start_date (date): Data when the pool should be activated.

        Returns:
            dict: Activation details.
        """
        params = dict(
            activationKey=regnum,
            startDate=str(start_date),
            systemName="genie",
            userName=username,
            vendor="REDHAT",
            webCustomerId=org_id,
        )

        response = self.api.post("/activate", json=params)
        response.raise_for_status()
        return response.json()



"""Ethel's API module

Provides access to all APIs, that are used by Ethel.
"""

from dataclasses import dataclass

from .candlepin import Candlepin
from .exceptions import EthelConnectionError, EthelError
from .subscription import ActivationV2, RegnumV5
from .terms import TermsV1
from .user import UserV1


@dataclass
class API:
    candlepin: Candlepin
    user: UserV1
    regnum: RegnumV5
    activation: ActivationV2
    terms: TermsV1


def initialize_apis(rest_host: str, candlepin_host: str) -> API:
    """Initialize APIs.

    Populate a API namedtuple with API clients for desired endpoints.

    Args:
        candlepin_host (str): Host of targed Candlepin
        rest_host (str): Base host for all REST APIs

    Returns:
        API: namedtuple containing all the clients.
    """
    return API(
        candlepin=Candlepin(candlepin_host),
        user=UserV1(rest_host),
        regnum=RegnumV5(rest_host),
        activation=ActivationV2(rest_host),
        terms=TermsV1(rest_host),
    )


__all__ = ("API", "initialize_apis", "EthelError", "EthelConnectionError")



from datetime import date, datetime, timedelta
from typing import Any, Callable, Optional, Union


def parse_date(value: Union[datetime, date, str] = None) -> date:
    """Date mapping.

    Map to date from string, datetime or date.

    Examples:
    >>> parse_date('yesterday')
    datetime.date(2020, 2, 5)
    >>> parse_date('today')
    datetime.date(2020, 2, 6)
    >>> parse_date('tomorrow')
    datetime.date(2020, 2, 7)
    >>> parse_date('2020-02-08')
    datetime.date(2020, 2, 8)
    >>> parse_date(datetime.date(2020, 2, 9))
    datetime.date(2020, 2, 9)
    >>> parse_date(datetime.datetime(2020, 2, 10))
    datetime.date(2020, 2, 10)

    Args:
        value (Union[datetime, date, str], optional): Parsed value. Defaults to None.

    Returns:
        date: Targed date object.
    """
    # pylint: disable=too-many-return-statements

    if not value:
        return datetime.today().date()

    if isinstance(value, datetime):
        return value.date()

    if isinstance(value, date):
        return value

    if value.lower() == "today":
        return date.today()

    if value.lower() == "yesterday":
        return date.today() - timedelta(days=1)

    if value.lower() == "tomorrow":
        return date.today() + timedelta(days=1)

    return date.fromisoformat(value)


def parse_duration(value: Union[int, timedelta] = timedelta(days=365)) -> timedelta:
    """Duration mapping.

    Parse a a timedelta or integer into a valid timedelta.

    Examples:
    >>> parse_duration()
    datetime.timedelta(days=365)
    >>> parse_duration(10)
    datetime.timedelta(days=10)
    >>> parse_duration(datetime.timedelta(days=10))
    datetime.timedelta(days=10)

    Args:
        value (Union[int, timedelta], optional): Parsed value. Defaults to a ~year.

    Raises:
        ValueError: Raised, if the duration is shorter than a day.

    Returns:
        timedelta: Target duration.
    """
    if isinstance(value, timedelta):
        if value < timedelta(days=1):
            raise ValueError("Duration must be at least a day")
        return value

    if not isinstance(value, int):
        raise ValueError("Supported datatype Union[int, timedata].")

    if value < 1:
        raise ValueError("Duration must be at least a day")
    return timedelta(days=value)


def apply_mapping(source_dict: dict, mapping: Union[Callable[[dict], Any], str]) -> Any:
    """Apply mapping to a dictionary.

    Apply a mapping function to a dictionary or query for a key directly.

    Args:
        source_dict (dict): Object of interest, where to look for data.
        mapping (Union[Callable[[dict], Any], str]): If string, it queries for the value
        in source_dict. If Callable, it calls this mapping function with source_dict as
        a argument.

    Returns:
        Any: Desired value.
    """
    if callable(mapping):
        return mapping(source_dict)
    return source_dict.get(mapping)


def get_instance_multiplier(source_dict: dict) -> Optional[int]:
    """Mapping function to get an instance multiplier.

    Args:
        source_dict (dict): Data source.

    Returns:
        int: Instance multiplier.
    """
    # "productAttributes" are encoded as {"name": ..., "value": ...} dicts
    # We're looking for value where x["name"] == "instance_multiplier"
    multiplier_gen = (
        int(a["value"])
        for a in source_dict.get("productAttributes", [])
        if a["name"] == "instance_multiplier"
    )

    return next(multiplier_gen, None)


def get_quantity(source_dict: dict) -> Union[float, str]:
    """Quantity mapping function.

    Args:
        source_dict (dict): Data source.

    Returns:
        Union[float, str]: Quantity.
    """
    quantity = source_dict["quantity"]
    if quantity < 0:
        return "unlimited"
    if quantity == 0:
        return 0

    instance_multiplier = get_instance_multiplier(source_dict) or 1

    return quantity / source_dict.get("multiplier", 1) / instance_multiplier



import os

from .base import APIBase
from .exceptions import raises_from_candlepin as raises_ethel_exception

ADMIN_AUTH = (
    os.getenv("CANDLEPIN_USERNAME", "candlepin_admin"),
    os.getenv("CANDLEPIN_PASSWORD", "candlepin_admin"),
)


class Candlepin(APIBase):
    def __init__(self, api_host: str) -> None:
        """Candlepin API

        Access the Candlepin API for subscription pool management.

        Args:
            api_host (str): Base API host
        """
        super().__init__(f"http://{api_host}/candlepin")

    def refresh(self, org_id: int) -> dict:
        """Force a Candlepin refresh.

        Forces Candlepin to synchronize EBS subscriptions to internal database.
        Requires admin access.

        Args:
            org_id (int): Organization ID of the user

        Returns:
            dict: Refresh job details
        """
        response = self.api.put(
            f"/owners/{org_id}/subscriptions",
            params=dict(auto_create_owner=True),
            auth=ADMIN_AUTH,
        )
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def get_job(self, job_id: str) -> dict:
        """Get job details.

        Args:
            job_id (str): Job ID

        Returns:
            dict: Job details
        """
        response = self.api.get(f"/jobs/{job_id}", auth=ADMIN_AUTH,)
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def get_owners(self, username: str, password: str) -> list:
        """Get list of owners.

        Args:
            username (str): Account's username.
            password (str): Account's password.

        Returns:
            list: List of account owners. Should contain 1 owner only.
        """
        response = self.api.get(f"/users/{username}/owners", auth=(username, password),)
        response.raise_for_status()
        return response.json()

    def delete_owner(self, username: str, password: str, owner_id: int) -> None:
        # pylint: disable=missing-function-docstring
        raise NotImplementedError

    @raises_ethel_exception
    def get_pools(
        self, username: str, password: str, owner_id: int, future: bool = False
    ) -> list:
        """Get list of subscription pools.

        Args:
            username (str): Account's username.
            password (str): Account's password.
            owner_id (int): Account's owner ID.
            future (bool, optional): List also subscription pools available in future.
                Defaults to False.

        Returns:
            list: List of pools available to the account.
        """
        response = self.api.get(
            f"/owners/{owner_id}/pools",
            params=dict(listall=True, add_future=future),
            auth=(username, password),
        )
        response.raise_for_status()
        return response.json()

    def delete_pool(self, username: str, password: str, pool_id: int) -> None:
        # pylint: disable=missing-function-docstring
        raise NotImplementedError

    def delete_user(self, username: str, password: str) -> None:
        # pylint: disable=missing-function-docstring
        raise NotImplementedError



import os

import yaml
from jinja2 import Template as Jinja2Template


class Template:
    def __init__(self, filename: str):
        """YAML Jinja2 template for request payloads.

        Loads a local YAML and converts it into a Jinja2 template. That makes it
        available for further processing via Jinja2 templating.

        Args:
            filename (str): YAML file name to load.
        """
        base = os.path.dirname(os.path.abspath(__file__))
        with open(f"{base}/{filename}", "r") as template:
            content = template.read()
        self.template = Jinja2Template(content)

    def render(self, **kwargs) -> dict:
        """Render the template.

        Pass variables to Jinja rendering function.

        Returns:
            dict: Rendered YAML document as a Python dict
        """
        render = self.template.render(**kwargs)
        return yaml.safe_load(render)



from datetime import date, datetime, timedelta
from typing import List, Union

from .api import API
from .utils import (apply_mapping, get_instance_multiplier, get_quantity,
                    parse_date, parse_duration)


class Account:
    def __init__(
        self,
        api: API,
        username: str,
        password: str,
        first_name: str = None,
        last_name: str = None,
        email: str = None,
        create_owners: bool = True,
        accept_terms: bool = True,
    ) -> None:
        """
        New account.

        Create new account.

        Args:
            api (API): API data structure instance.
            username (str): Username to use for the new account.
            password (str): Password to use for the new account.
            first_name (str, optional): Specify the first name. Defaults to None.
            last_name (str, optional): Specify the last name. Defaults to None.
            email (str, optional): Specify the email. Defaults to None.
            create_owners (bool, optional): Perform a Candlepin refresh to populate
                candlepin owners account. Defaults to True.
            accept_terms (bool, optional): Activate the account by acception Terms and
                Conditions. Defaults to True.
        """
        self.username = username
        self.password = password
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.api = api

        self._org_id: int = None  # type: ignore
        self._owner_id: int = None  # type: ignore
        self._latest_refresh_job_id: str = None  # type: ignore
        self.orders: List[dict] = []
        self.activations: List[dict] = []

        if self.does_exist():
            self.login()
            return

        self.create()

        if create_owners:
            self.start_refresh()

        if accept_terms:
            self.accept_all_terms()

    @property
    def org_id(self) -> int:
        """Organization ID.

        Get the account's organization ID if not cached already.

        Returns:
            int: Organization ID
        """
        if self._org_id is None:
            account_list = self.api.user.login(self.username)
            account = account_list[0] if account_list else {}
            self._org_id = account.get("orgId")

        return self._org_id

    @property
    def owner_id(self) -> int:
        """Candlepin owner account ID.

        Get the Candlepin Owner account ID if not cached already.

        Returns:
            int: Owner ID
        """
        if not self._owner_id:
            owners = self.api.candlepin.get_owners(self.username, self.password)
            if not len(owners) == 1:
                raise IndexError('A single owner is expected', owners)
            self._owner_id = int(owners[0].get("key"))

        return self._owner_id

    def does_exist(self) -> bool:
        """Check if account already exists.

        Check if there's only one account with the same username

        Returns:
            bool: True if the account exists
        """
        return len(self.api.user.login(self.username)) == 1

    def create(self) -> bool:
        """Create the user account via API request.

        Returns:
            bool: True if account is created successfully
        """
        account_id = self.api.user.create(
            username=self.username,
            password=self.password,
            first_name=self.first_name,
            last_name=self.last_name,
            email=self.email,
        )
        return isinstance(account_id, int)

    def login(self) -> bool:
        """Login to Candlepin using own credentials.

        Raises:
            EthelError: In case credentials are not valid.

        Returns:
            bool: True if everything is OK.
        """
        self.api.candlepin.get_owners(self.username, self.password)
        return True

    POOL_ATTRIBUTES_MAPPING = dict(
        pool_id="id",
        sku_id="productId",
        product_name="productName",
        start_date="startDate",
        end_date="endDate",
        muiltiplier="multiplier",
        quantity=get_quantity,
        instance_multiplier=get_instance_multiplier,
    )

    def list_pools(  # pylint: disable=dangerous-default-value
        self, future: bool = False, filter_attributes: dict = POOL_ATTRIBUTES_MAPPING
    ) -> list:
        """List all subscriptions to this account.

        Args:
            future (bool, optional): List also subscription pools available in future.
                Defaults to False.
            filter_attributes (dict, optional): This parameter allows you to modify what
                values are parsed out of the API response. If set to empty dict, it
                returns raw API respose, as is. Supported format:
                    key: Target attribute name
                    value: Either a key of the API response dict or a callable
                Defaults to Account.POOL_ATTRIBUTES_MAPPING

        Returns:
            list: List of all subscriptions.
        """
        raw_pools = self.api.candlepin.get_pools(
            self.username, self.password, self.owner_id, future=future
        )
        if not filter_attributes:
            return raw_pools

        pools = [
            {k: apply_mapping(pool, v) for k, v in filter_attributes.items()}
            for pool in raw_pools
        ]

        return pools

    def start_refresh(self) -> None:
        """Requests a Candlepin refresh job.

        Requests Candlepin to propagate EBS changes to it's internal database, making
        new users and new subscriptions available.
        """
        job_id = self.api.candlepin.refresh(self.org_id).get("id")
        if job_id:
            self._latest_refresh_job_id = job_id

    def get_refresh_status(self) -> str:
        """Check refresh job status.

        If a refresh job was previously started, this method allows to query
        Candlepin for it's current state.

        Returns:
            str: Job status
        """
        if not self._latest_refresh_job_id:
            return "UNKNOWN"

        job = self.api.candlepin.get_job(self._latest_refresh_job_id)
        return job.get("state", "UNKNOWN")

    def subscribe(
        self,
        sku_id: str,
        quantity: int = 1,
        start_date: Union[datetime, date, str] = None,
        duration: Union[timedelta, int] = 365,
    ) -> int:
        """Create subscription to a product.

        Order a SKU and activate the subscription pool, so it's ready to be consumed.
        Subscription availability timeframe can be adjusted by modifying a start_date
        and duration. However, APIs consider duration in days only. Also not many
        subscriptions are available to be created with a duration set to just few days.
        Therefore, it's advised, if you want to create a subscription that lasts just
        for a few days, rather antidate the start_date.

        Args:
            sku_id (str): SKU identifier.
            quantity (int, optional): SKU quantity. Defaults to 1.
            start_date (Union[datetime, date, str], optional): Start date of the
                subscription. See ethel.utils.parse_date for all accepted values.
                Defaults to None.
            duration (Union[timedelta, int], optional): Subscription duration. See
                ethel.utils.parse_duration for all accepted values. Defaults to 365.

        Returns:
            int: Subscription ID
        """
        start_date = parse_date(start_date)
        duration = parse_duration(duration)

        order = self.api.regnum.order(
            self.username, sku_id, quantity, start_date, duration
        )
        self.orders.append(order)
        registration_num = order["regNumbers"][0][0]["regNumber"]

        activation = self.api.activation.activate(
            self.username, self.org_id, registration_num, start_date
        )
        self.activations.append(activation)
        return activation["id"]

    def accept_all_terms(self, optional: bool = False) -> None:
        """Accept all Terms and Conditions.

        Lists and accepts all required (and optional) Terms and Conditions.

        Args:
            optional (bool): Accept also optional Terms. Dafaluts to False
        """
        if optional:
            all_terms = self.api.terms.get_all_terms(self.username)
        else:
            all_terms = self.api.terms.get_required_terms(self.username)

        # Accept terms using a randomly selected PDF translation
        for terms in all_terms:
            translations = terms.get("translations", [])
            if len(translations) <= 0:
                continue

            pdf_id = translations[0].get("termsPdfId")
            self.api.terms.accept_terms(self.username, pdf_id)



from .base import CERT, APIBase
from .exceptions import raises_from_ebs as raises_ethel_exception
from .utils import Template


class UserV1(APIBase):
    CREATE_PAYLOAD_TEMPLATE = Template("payloads/user.yml")

    def __init__(self, api_host: str) -> None:
        """User API

        Access the old API for user management.

        Args:
            api_host (str): Base API host
        """
        super().__init__(f"https://user.{api_host}/svcrest/user/v3", cert=CERT)

    @raises_ethel_exception
    def create(
        self,
        username: str,
        password: str,
        first_name: str = None,
        last_name: str = None,
        email: str = None,
    ) -> int:
        """Create user.

        Make a call to the /create endpoint with payload.

        Args:
            payload (dict): Json-like payload for the endpoint
            username (str): Acccount's username.
            password (str): Account's password.
            first_name (str, optional): First name, left blank if None.
            last_name (str, optional): Last name, left blank if None.
            email (str, optional): Account's email. Left blank if None.

        Returns:
            int: If successful, an account id is returned
        """
        payload = self.CREATE_PAYLOAD_TEMPLATE.render(
            username=username,
            password=password,
            first_name=first_name,
            last_name=last_name,
            email=email,
        )

        response = self.api.post("/create", json=payload)
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def login(self, username: str) -> list:
        """Login user.

        Make a call to the /login endpoint with the nonstandard call.

        Args:
            username (str): Username used for the account

        Returns:
            list: List of users matching the username
        """
        response = self.api.get(f"/login={username}")
        response.raise_for_status()
        return response.json()


class UserV2(APIBase):
    CREATE_PAYLOAD_TEMPLATE = Template("payloads/user_v2.yml")

    def __init__(self, api_host: str) -> None:
        """User API

        Access the old API for user management.

        Args:
            api_host (str): Base API host
        """
        super().__init__(f"https://user.{api_host}/v2", cert=CERT)

    @raises_ethel_exception
    def create(
        self,
        username: str,
        password: str,
        first_name: str = None,
        last_name: str = None,
        email: str = None,
    ) -> int:
        """Create user.

        Make a call to the /create endpoint with payload.

        Args:
            payload (dict): Json-like payload for the endpoint
            username (str): Acccount's username.
            password (str): Account's password.
            first_name (str, optional): First name, left blank if None.
            last_name (str, optional): Last name, left blank if None.
            email (str, optional): Account's email. Left blank if None.

        Returns:
            int: If successful, an account id is returned
        """
        payload = self.CREATE_PAYLOAD_TEMPLATE.render(
            username=username,
            password=password,
            first_name=first_name,
            last_name=last_name,
            email=email,
        )

        response = self.api.post("/createUser", json=payload)
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def login(self, username: str) -> list:
        """Login user.

        Make a call to the /findUser endpoint with a payload searching for an exact
        match on username.

        Args:
            username (str): Username used for the account

        Returns:
            list: List of users matching the username
        """
        payload = dict(
            by=dict(authentication=dict(provider="Red Hat", principal=username))
        )
        response = self.api.post("/findUser", json=payload)
        response.raise_for_status()
        return response.json()



import atexit
import os
from typing import Tuple

import requests

CERT = (os.getenv("EBS_CERT_PUBLIC", ""), os.getenv("EBS_CERT_KEY", ""))


class APISession(requests.Session):
    def __init__(
        self, api_base_url: str, cert: Tuple[str, str] = None, verify: bool = False,
    ) -> None:
        """API session with base url.

        A requests.Session with a base url and certificates settings available. It allows
        user to set a common host for all API requests to lower any confusion.

        Args:
            api_base_url (str): Base URL (API host)
            cert (tuple, optional): SSL client certificate to use for all requests.
                Defaults to None.
            verify (bool, optional): SSl CA verification mode. Defaults to False.
        """
        super().__init__()
        atexit.register(self.close)

        self.cert = cert
        self.verify = verify
        self.api_base_url = api_base_url.rstrip("/")

        # Inject api_base_url to the url param of every request.
        def override(parent_method):
            def wrapper(url, *args, **kwargs):
                return parent_method(self.api_base_url + url, *args, **kwargs)

            return wrapper

        parent = super()
        for method in ("get", "options", "head", "post", "put", "patch", "delete"):
            parent_method = getattr(parent, method)
            setattr(self, method, override(parent_method))


class APIBase:
    def __init__(
        self, api_base_url: str, cert: Tuple[str, str] = None, verify: bool = False
    ) -> None:
        """API Base.

        Args:
            api_base_url (str): Base API URL. Passed to APISession.
            cert (tuple, optional): SSL client certificate to use for all requests.
                Passed to APISession. Defaults to None.
            verify (bool, optional): SSl CA verification mode. Passed to APISession.
                Defaults to False.
        """
        self.api = APISession(api_base_url, cert, verify)

    def __repr__(self):
        return "{0}(api_base_url={1})".format(
            self.__class__.__name__, self.api.api_base_url
        )



from functools import wraps

from requests import ConnectionError as RequestsConnectionError
from requests import HTTPError, Timeout

from .base import APIBase


class EthelError(IOError):
    def __init__(
        self,
        message: str,
        *args,
        raw_error: HTTPError,
        status_code: str = None,
        source: APIBase = None,
        exception_type: str = None,
        message_type: str = None,
        uuid: str = None,
    ):
        """Ethel Exception.

        Translation layer between raw HTTPError and it's payload.

        Args:
            message (str): Error message.
            raw_error (HTTPError): Raw requests.HTTPError object for further detail.
            status_code (str, optional): HTTP status code. Defaults to None.
            source (APIBase, optional): API which raised this exception.
                Defaults to None.
            exception_type (str, optional): Exception type as a string if available.
                Defaults to None.
            message_type (str, optional): Message type as a string if available.
                Defaults to None.
            uuid (str, optional): UUID of the operation if available. Defaults to None.
        """
        self.message = message
        self.status_code = status_code
        self.raw_error = raw_error
        self.source = source
        self.exception_type = exception_type
        self.message_type = message_type
        self.uuid = uuid

        super().__init__(*args)

    def __str__(self):
        status = f" Status code: {self.status_code}." if self.status_code else ""
        ex_type = f" Type: {self.exception_type}." if self.exception_type else ""

        request = self.raw_error.request
        call = f" Call(method={request.method}, url={request.url})."
        return f"From: {self.source}. Reason: {self.message}.{status}{ex_type}{call}"


class EthelConnectionError(RequestsConnectionError, Timeout):
    """Ethel's shorthand for easier catching of RequestsConnectionError and Timeout."""


def raises_from_candlepin(func):
    """Map Candlepin exception response JSON to EthelError.

    Args:
        func ([type]): Decorated function.

    Raises:
        EthelError

    Returns:
        Return value of func.
    """

    @wraps(func)
    def wrapper(self: APIBase, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except HTTPError as http_error:
            try:
                response = http_error.response.json()
            except ValueError as value_error:
                raise EthelError(
                    message=str(value_error),
                    status_code=http_error.response.status_code,
                    exception_type="JSONDecodeError",
                    source=self,
                    raw_error=http_error,
                )

            raise EthelError(
                message=response["displayMessage"],
                status_code=http_error.response.status_code,
                source=self,
                uuid=response["requestUuid"],
                raw_error=http_error,
            )

    return wrapper


def raises_from_ebs(func):
    """Map EBS exception response JSON to EthelError.

    Args:
        func ([type]): Decorated function.

    Raises:
        EthelError

    Returns:
        Return value of func.
    """

    @wraps(func)
    def wrapper(self: APIBase, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except HTTPError as http_error:
            try:
                response = http_error.response.json()
            except ValueError as value_error:
                # NOTE: EBS sometimes decides to rather return and empty response
                # pylint: disable=no-member
                message = (
                    "API returned an empty response"
                    if value_error.pos == 0  # type: ignore
                    else str(value_error)
                )

                raise EthelError(
                    message=message,
                    status_code=http_error.response.status_code,
                    exception_type="JSONDecodeError",
                    source=self,
                    raw_error=http_error,
                )

            raise EthelError(
                message=response["message"],
                message_type=response["msgName"],
                status_code=http_error.response.status_code,
                exception_type=response["type"],
                source=self,
                raw_error=http_error,
            )

    return wrapper



from .base import CERT, APIBase
from .exceptions import raises_from_ebs as raises_ethel_exception


class TermsV1(APIBase):
    def __init__(self, api_host: str) -> None:
        """Terms API

        Access the API for Terms and Conditions management.

        Args:
            api_host (str): Base API host
        """
        super().__init__(
            f"https://terms.{api_host}/svcrest/terms/presentation", cert=CERT
        )

    @raises_ethel_exception
    def get_required_terms(self, username: str) -> list:
        """Get all Terms and Conditions the user is required to accept.

        Args:
            username (str): Account's username.

        Returns:
            list: List of events
        """
        params = dict(login=username, event="attachSubscription", site="candlepin")
        response = self.api.get("/required", params=params)
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def get_all_terms(self, username: str) -> list:
        """Get all Terms and Conditions for the user.

        List all, required as well as optional terms.

        Args:
            username (str): Account's username.

        Returns:
            list: List of events
        """
        params = dict(login=username, event="attachSubscription", site="candlepin")
        response = self.api.get("/available", params=params)
        response.raise_for_status()
        return response.json()

    @raises_ethel_exception
    def accept_terms(self, username: str, terms_id: int) -> bool:
        """Accept a Terms and Conditions document.

        Accepts single document based on its PDF ID.

        Args:
            username (str): Account's username.
            terms_id (int): PDF ID of the Terms and Conditions.

        Returns:
            bool: True if successfull
        """
        params = dict(
            login=username, pdfid=terms_id, acknowledgedcode="accepted", site="candlepin"
        )
        response = self.api.put("/ackterms", params=params)
        response.raise_for_status()
        return response.content == b""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ethel', 'ethel.api']

package_data = \
{'': ['*'], 'ethel.api': ['payloads/*']}

install_requires = \
['Jinja2', 'pyyaml', 'requests']

setup_kwargs = {
    'name': 'ethel',
    'version': '3.0.0rc2',
    'description': 'Account Management Tool for Testing',
    'long_description': '# Ethel\n\nEthel is a tool that allows you to manipulate with accounts in Stage, QA or any other environment, buy Subscriptions, activate them and do other chores.\n\n## Instalation\n\n```sh\npip install ethel\n```\n\n## Usage\n\n### Prerequisites\n\nBefore you can use Ethel, you have to set up your environment accordingly:\n\n```sh\n# Required:\nEBS_CERT_PUBLIC=<PATH_TO_EBS_SSL_CLIENT_CERTIFICATE>\nEBS_CERT_KEY=<PATH_TO_EBS_SSL_CLIENT_CERTIFICATE_KEY>\n\n# Optional:\nCANDLEPIN_USERNAME=<CANDLEPIN_SUPERUSER_USERNAME>  # Defaults to "candlepin_admin"\nCANDLEPIN_PASSWORD=<CANDLEPIN_SUPERUSER_PASSWORD>  # Defaults to "candlepin_admin"\n```\n\n### Basic usage\n\nThe most basic usecase is to simply create an account and subscribe it to a product.\n\n```python\n>>> from ethel import Ethel\n\n>>> ethel = Ethel.stage()\n\n>>> account = ethel.create_account(\'some_fancy_username\', \'not_so_secret_password\')\n\n>>> account.subscribe(\'product_sku\')\n```\n\nEthel provides access to Stage and QA environments via `Ethel.stage()` and `Ethel.qa()` class methods.\n\n### Advanced usage\n\nEthel, by default, processes everything for you when the account is being created. Also, if account with the same username already exists, Ethel verifies your credentials and returns you the already existant account entry.\n\nWhen account is being created, Ethel ensures that it\'s Candlepins owners are created (forces Candlepin refresh) and accepts all required Terms and Conditions. To disable this behavior, you can pass `create_owners=False` and/or `accept_terms=False` when creating account and do it yourself later:\n\n```python\n>>> account = ethel.create_account(\n... "some_fancy_username",\n... "not_so_secret_password",\n... first_name="Gretchen",\n... last_name="SomeOldSurname",\n... email="grandmas.dont.have@emails.com",\n... create_owners=False,\n... accept_terms=False,\n... )\n\n>>> account.start_refresh()\n\n>>> account.get_refresh_status()\n\'RUNNING\'\n\n>>> account.get_refresh_status()\n\'FINISHED\'\n\n>>> account.accept_all_terms(optional=True)  # Accepts also the optional Terms and Conditions\n```\n\nYou can also specify more details about your desired subscription when asking Ethel to subscribe it you your account:\n\n```python\n>>> account.subscribe(\n... \'product_sku\',\n... quantity=42,\n... start_date=\'yesterday\',  # accepts \'yesterday\', \'today\', \'tomorrow\', ISO date \'YYYY-MM-DD\' or datetime.date\n... duration=365 # accepts integer that means \'days\' or datetime.timedelta\n... )\n```\n\n### Errors and Exceptions\n\nIf an exception is returned to Ethel from either Candlepin or the EBS rest API services, they are unified and interfaced as an `EthelError`. Depending on the exact API that raised the exception, the level of detail varies. Following properties are stored:\n\n- `message` - Should contain the reason why the exception occurred.\n- `message_type` - Class of the exception reason description (different to `exception_type`).\n- `exception_type` - Type of exception (class name).\n- `status_code` - HTTP status code (EBS likes to return HTTP_500 for any reason).\n- `source` - API that caused the exception.\n- `uuid` - Exception\'s tracking code for Candlepin API.\n- `raw_error` - `HTTPError` object providing access to `PreparedRequest` and `Response` directly.\n\nEthel does not handle `requests.ConnectionError` and `requests.Timeout`. **For convenience a shorthand `EthelConnectionError` is provided.** Ethel doesn\'t retry any request, it\'s up to user to handle this behavior.\n\n```python\n>>> account = ethel.create_account(\'USERNAME\', \'WRONG_PASSWORD\')\nEthelError: From: Candlepin(api_base_url=<CANDLEPIN_URL_FOR_THIS_ENV>). Reason: Invalid user credentials. Status code: 401. Call(method=GET, url=<CANDLEPIN_URL_FOR_THIS_ENV>)/users/<USERNAME>/owners).\n\n>>> try:\n...     account = ethel.create_account(\'<USERNAME>\', \'<WRONG_PASSWORD>\')\n... except EthelError as e:\n...     logging.error("Status Code %d, Message: %s, Tracking code %s" % (e.status_code, e.message, e.uuid))\n\nERROR:root:Status Code 401, Message: Invalid user credentials, Tracking code e7275480-2c59-42fa-a6db-b8df6a9dc323\n```\n\n### Additional properties\n\nEthel stores few important properties for each of your accounts that might be usefull to you:\n\n- Organization ID\n- Candlepin Owner ID\n- List of all Subscription orders done in this session\n- List of all Activation orders done in this session\n\n```python\n>>> account.org_id\n123456789\n\n>>> account.owner_id\n987654321\n\n>>> account.orders\n[...]\n\n>>> account.activations\n[...]\n```\n\n### View Account\n\nAnd since Ethel allows you to have a Black Friday for subscriptions everyday, you may want to look up all the things you\'ve bought:\n\n```python\n>>> account.list_pools()  # You can pass future=True argument to list also pools valid in future\n[\n    {\n        \'pool_id\': \'<USE_THIS_TO_SUBSCRIBE>\',\n        \'sku_id\': \'<PRODUCT_SKU_ID>\',\n        \'product_name\': \'Fancy Product Name\',\n        \'start_date\': \'2020-02-07T05:00:00+0000\',\n        \'end_date\': \'2021-02-07T04:59:59+0000\',\n        \'muiltiplier\': None,\n        \'quantity\': \'unlimited\',\n        \'instance_multiplier\': None\n    },\n    {\n        \'pool_id\': \'<ALSO_USE_THIS_TO_SUBSCRIBE>\',\n        \'sku_id\': \'<ANOTHER_PRODUCT_SKU_ID>\',\n        \'product_name\': \'Fancy Product Name\',\n        \'start_date\': \'2020-02-06T05:00:00+0000\',\n        \'end_date\': \'2021-02-06T04:59:59+0000\',\n        \'muiltiplier\': None,\n        \'quantity\': 1,\n        \'instance_multiplier\': 16\n    },\n    ...\n]\n```\n\n## Developer setup\n\nAfter cloning this repo, setup the local environment via [Poetry](https://python-poetry.org/):\n\n```sh\npoetry install\n```\n\n### Prepared tasks\n\n- `poetry run task lint` - runs [Mypy](http://mypy-lang.org/) and [Pylint](https://www.pylint.org/)\n- `poetry run task test` - runs [Pytest](https://docs.pytest.org/en/latest/) test suite\n',
    'author': 'Tomas Coufal',
    'author_email': 'tcoufal@redhat.com',
    'maintainer': 'Tomas Coufal',
    'maintainer_email': 'tcoufal@redhat.com',
    'url': 'https://github.com/tumido/ethel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .account import Account
from .api import initialize_apis

HOSTS = dict(
    stage=("stage.api.redhat.com", "candlepin.dist.stage.ext.phx2.redhat.com"),
    qa=("qa.api.redhat.com", "candlepin.corp.qa.redhat.com"),
)


class Ethel:
    def __init__(self, rest_host: str, candlepin_host: str):
        """Ethel.

        Args:
            candlepin_host (str): Host of targed Candlepin
            rest_host (str): Base host for all REST APIs
        """
        self.api = initialize_apis(rest_host, candlepin_host)

    @classmethod
    def stage(cls) -> "Ethel":
        """Returns Ethel instance for Stage environment."""
        return cls(*HOSTS["stage"])

    @classmethod
    def qa(cls) -> "Ethel":  # pylint: disable=invalid-name
        """Returns Ethel instance for QA environment."""
        return cls(*HOSTS["qa"])

    def create_account(self, *args, **kwargs) -> Account:
        """Creates a new account.

        New account is created if it doesn't exist. If it does, it tries to
        log in into the account.

        Returns:
            Account: Account object.
        """
        return Account(self.api, *args, **kwargs)
